/*
** -- Local Includes --
*/

#include "qtchess_comm.h"

extern qtchess *chess;
extern qtchess_gui *gui;
extern qtchess_comm *comm;
extern QApplication *qapp;
static QMutex mutex;
static QSocketDevice *sock = NULL;

void qtchess_comm_thread::run()
{
  int sock_fd = -1;
  char buffer[BUFFER_SIZE];
  Q_LONG n = 0;
  QSocketDevice *listener = NULL;

  while(!interrupted)
    {
      if((listener = comm->getListener()) == NULL)
	{
	  QThread::sleep(1);
	  continue;
	}

      /*
      ** Allow only one connection.
      */

      if(sock == NULL)
	{
	  if((sock_fd = listener->accept()) < 0)
	    {
	      QThread::sleep(1);
	      continue;
	    }

	  sock = new QSocketDevice(sock_fd, QSocketDevice::Stream);

	  if(interrupted)
	    break;

	  if(sock == NULL)
	    {
	      QThread::sleep(1);
	      continue;
	    }

	  gui->notifyConnection(sock->peerAddress().toString());
	  listener->close();

	  /*
	  ** Lock.
	  */

	  mutex.lock();

	  if(chess->getFirst() == -1)
	    {
	      chess->setTurn(THEIR_TURN);
	      chess->setFirst(THEY_ARE_FIRST);
	      chess->setMyColor(BLACK);
	    }

	  /*
	  ** Unlock.
	  */

	  mutex.unlock();

	  /*
	  ** Set to blocking.
	  */

#ifdef _DEBUG_
	  (void) fprintf(stderr, "Connected.\n");
#endif

	  do
	    {
	      if(interrupted)
		break;

	      (void) memset(buffer, 0, sizeof(buffer));

	      if((n = sock->readBlock(buffer, sizeof(buffer))) > 0)
		if(QString(buffer).contains("\r\n"))
		  chess->updateBoard(buffer);

#ifdef _DEBUG_
	      (void) fprintf(stderr, "readBlock() returned %d bytes.\n", n);
#endif

	      if(!interrupted)
		QThread::sleep(1); // Prevent flooding.
	    }
	  while(n > 0);

	  if(interrupted)
	    break;

#ifdef _DEBUG_
	  (void) fprintf(stderr, "Finished reading.\n");
#endif

	  mutex.lock();
	  chess->setTurn(-1);
	  chess->setFirst(-1);
	  chess->setMyColor(-1);
	  comm->setListen();
	  mutex.unlock();
	  listener = comm->getListener();
	  comm->disconnectRemotely();
	  gui->showDisconnect();
	  sock->close();
	  delete sock;
	  sock = NULL;
	}

      if(!interrupted)
	QThread::sleep(1);
    }

  if(sock != NULL)
    {
      sock->close();
      delete sock;
      sock = NULL;
    }

#ifdef _DEBUG_
  (void) fprintf(stderr, "Exiting listening thread.\n");
#endif
}

void qtchess_comm::init(void)
{
  connected = FALSE;

  if(listening_sock != NULL)
    {
      listening_sock->close();
      delete listening_sock;
      listening_sock = NULL;
    }

  if((conn_thread = new qtchess_comm_thread()) == NULL)
    chess->quit("Memory allocation failure.", EXIT_FAILURE);

  /*
  ** Create a listener thread.
  */

  setListen();

  if(listening_sock != NULL)
    gui->getSetupDialog()->getHostField()->setText("127.0.0.1");

  /*
  ** Launch a thread to listen for incomming connections.
  ** Only one connection will be accepted. That's the plan.
  */

  conn_thread->start();
}

int qtchess_comm::isSet(void)
{
  if(connected)
    return TRUE;
  else
    return FALSE;
}

int qtchess_comm::isReady(void)
{
  if(isSet())
    {
      if(sock != NULL && send_sock != NULL)
	return TRUE;
      else
	return FALSE;
    }
  else
    return FALSE;
}

void qtchess_comm::setListen(void)
{
  if(listening_sock != NULL)
    {
      listening_sock->close();
      delete listening_sock;
      listening_sock = NULL;
    }

  /*
  ** Create a stream socket.
  */

  if((listening_sock = new QSocketDevice(QSocketDevice::Stream)) == NULL)
    chess->quit("socket() failure.", EXIT_FAILURE);

  /*
  ** Bind the socket to the port.
  */

  if(!listening_sock->bind(listening_sock->address(), listening_sock->port()))
    chess->quit("bind() failure.", EXIT_FAILURE);

  /*
  ** Save the port number.
  */

  qapp->lock();
  gui->getSetupDialog()->getPortField()->setText
    (QString::number(listening_sock->port()));
  qapp->unlock();

  /*
  ** Listen.
  */

  if(!listening_sock->listen(1))
    chess->quit("listen() failure.", EXIT_FAILURE);
}

QSocketDevice *qtchess_comm::getListener(void)
{
  return listening_sock;
}

void qtchess_comm::quit(void)
{
  /*
  ** Terminate all communications.
  */

  if(conn_thread != NULL)
    conn_thread->setInterrupted(TRUE);

  if(send_sock != NULL)
    send_sock->close();

#ifdef _DEBUG_
  else
    (void) fprintf(stderr, "send_sock is NULL\n");
#endif

  if(listening_sock != NULL)
    listening_sock->close();

#ifdef _DEBUG_
  else
    (void) fprintf(stderr, "listening_sock is NULL\n");
#endif

  if(conn_thread != NULL)
    conn_thread->wait();

  delete send_sock;
  send_sock = NULL;
  mutex.lock();
  delete listening_sock;
  listening_sock = NULL;
  mutex.unlock();

  setConnected(FALSE);
#ifdef _DEBUG_
  (void) fprintf(stderr, "Connection(s) terminating.\n");
#endif
}

void qtchess_comm::setConnected(const int connected_arg)
{
  this->connected = connected_arg;
}

void qtchess_comm::disconnectRemotely(void)
{
  if(send_sock != NULL)
    {
      send_sock->close();
      delete send_sock;
      send_sock = NULL;
    }

  chess->setTurn(-1);
  chess->setFirst(-1);
  chess->setMyColor(-1);
  setConnected(FALSE);
}

void qtchess_comm::connectRemotely(void)
{
  QString str1 = "", str2 = "";
  Q_UINT16 remotePort = 0;
  QHostAddress remoteAddress;

  str1 = gui->getSetupDialog()->getRHostField()->text();
  str1 = str1.stripWhiteSpace();
  str2 = gui->getSetupDialog()->getRPortField()->text();
  str2 = str2.stripWhiteSpace();
  remotePort = (Q_UINT16) atoi(str2);

  if((send_sock = new QSocketDevice(QSocketDevice::Stream)) == NULL)
    return;

  remoteAddress.setAddress(str1);

  if(!send_sock->connect(remoteAddress, remotePort))
    {
      send_sock->close();
      delete send_sock;
      send_sock = NULL;
      return;
    }

  setConnected(TRUE);
  mutex.lock();

  if(chess->getFirst() == -1)
    {
      chess->setTurn(MY_TURN);
      chess->setFirst(I_AM_FIRST);
      chess->setMyColor(WHITE);
    }

  mutex.unlock();
}

void qtchess_comm::sendMove(const struct move_s current_move)
{
  int i = 0, j = 0;
  char buffer[BUFFER_SIZE];
  char numstr[32];

  if(send_sock == NULL)
    return;

  /*
  ** Copy the structure.
  */

  (void) memset(buffer, 0, sizeof(buffer));
  (void) snprintf(buffer, sizeof(buffer),
		  "%d %d %d %d %d %d %d %d %d %d %d %d ",
		  current_move.x1,
		  current_move.x2,
		  current_move.y1,
		  current_move.y2,
		  current_move.r_x1,
		  current_move.r_x2,
		  current_move.r_y1,
		  current_move.r_y2,
		  current_move.piece,
		  current_move.rook,
		  current_move.promoted,
		  current_move.pawn_2);

  for(i = 0; i < NSQUARES; i++)
    for(j = 0; j < NSQUARES; j++)
      {
	(void) memset(numstr, 0, sizeof(numstr));
	(void) snprintf(numstr, sizeof(numstr), "%d ",
			current_move.board[i][j]);
	(void) strcat(buffer, numstr);
      }

  /*
  ** Remove the extra space.
  */

  buffer[strlen(buffer) - 1] = '\0';
  (void) strcat(buffer, "\r\n");

  if(send_sock->writeBlock(buffer, (Q_ULONG) strlen(buffer)) == -1)
    gui->showErrorMsg("Move could not be delivered.");
  else
    chess->setTurn(THEIR_TURN);
}

qtchess_comm::qtchess_comm(void)
{
  sock = NULL;
  send_sock = NULL;
  conn_thread = NULL;
  listening_sock = NULL;
};
